@@ -1,4 +1,4 @@
-VERSION = '0.2'
+VERSION = '0.3'
 
 
 from .tokenizer import parse_component_value_list
@@ -6,4 +6,4 @@ from .parser import (
     parse_one_component_value, parse_one_declaration, parse_declaration_list,
     parse_one_rule, parse_rule_list, parse_stylesheet)
 from .bytes import parse_stylesheet_bytes
-from .utils import ascii_lower
+from .serializer import serialize, serialize_identifier
