@@ -3,33 +3,52 @@
 
 Data structures for the CSS abstract syntax tree.
 
-Differences with css-syntax:
-
-* :class:`LiteralToken` regroups
-  colon ``:``, semicolon ``;``, comma ``,``, cdo ``<!--``, cdc ``-->``,
-  include-match ``~=``, dash-match ``|=``,
-  prefix-match ``^=``, suffix-match ``$=``, substring-match ``*=``,
-  and delim tokens.
-  (Delim is any single character not matched by another token.)
-
 """
 
 from __future__ import unicode_literals
 
-from .utils import ascii_lower
+from webencodings import ascii_lower
+
+from .serializer import (serialize_identifier, serialize_name,
+                         serialize_string_value, _serialize_to)
 
 
 class Node(object):
-    """Base class for all tokens.
+    """Every node type inherits from this class,
+    which is never instantiated directly.
+
+    .. attribute:: type
+
+        Each child class has a :attr:`type` class attribute
+        with an unique string value.
+        This allows checking for the node type with code like:
+
+        .. code-block:: python
+
+            if node.type == 'whitespace':
+
+        instead of the more verbose:
+
+        .. code-block:: python
+
+            from tinycss2.ast import WhitespaceToken
+            if isinstance(node, WhitespaceToken):
+
+    Every node also has these attributes and methods,
+    which are not repeated for brevity:
 
     .. attribute:: source_line
 
         The line number of the start of the node in the CSS source.
+        Starts at 1.
 
     .. attribute:: source_column
 
-        The column number within :attr:`line` of the start of the node
+        The column number within :attr:`source_line` of the start of the node
         in the CSS source.
+        Starts at 1.
+
+    .. automethod:: serialize
 
     """
     __slots__ = ['source_line', 'source_column']
@@ -38,9 +57,33 @@ class Node(object):
         self.source_line = source_line
         self.source_column = source_column
 
+    def __repr__(self):
+        return self.repr_format.format(self=self)
+
+    def serialize(self):
+        """Serialize this node to CSS syntax and return an Unicode string."""
+        chuncks = []
+        self._serialize_to(chuncks.append)
+        return ''.join(chuncks)
+
+    def _serialize_to(self, write):
+        """Serialize this node to CSS syntax, writing chuncks as Unicode string
+        by calling the provided :obj:`write` callback.
+
+        """
+        raise NotImplementedError
+
 
 class ParseError(Node):
-    """A syntax error of some sort.
+    """A syntax error of some sort. May occur anywhere in the tree.
+
+    Syntax errors are not fatal in the parser
+    to allow for different error handling behaviors.
+    For example, an error in a Selector list makes the whole rule invalid,
+    but an error in a Media Query list only replaces one comma-separated query
+    with ``not all``.
+
+    .. autoattribute:: type
 
     .. attribute:: kind
 
@@ -55,46 +98,104 @@ class ParseError(Node):
     """
     __slots__ = ['kind', 'message']
     type = 'error'
+    repr_format = '<{self.__class__.__name__} {self.kind}>'
 
     def __init__(self, line, column, kind, message):
         Node.__init__(self, line, column)
         self.kind = kind
         self.message = message
 
+    def _serialize_to(self, write):
+        if self.kind == 'bad-string':
+            write('"[bad string]\n')
+        elif self.kind == 'bad-url':
+            write('url([bad url])')
+        elif self.kind in ')]}':
+            write(self.kind)
+        else:
+            raise TypeError('Can not serialize %r' % self)
+
 
 class Comment(Node):
-    """A CSS comment."""
+    """A CSS comment.
+
+    By default, comments are ignored
+    and :class:`Comment` objects are not created.
+    This can be changed by passing ``preserve_comments=True``
+    to :func:`~tinycss2.parse_component_value_list`
+
+    .. autoattribute:: type
+    .. attribute:: value
+
+        The content of the comment, between ``/*`` and ``*/``.
+
+    """
     __slots__ = ['value']
     type = 'comment'
+    repr_format = '<{self.__class__.__name__} {self.value:r}>'
 
     def __init__(self, line, column, value):
         Node.__init__(self, line, column)
         self.value = value
 
+    def _serialize_to(self, write):
+        write('/*')
+        write(self.value)
+        write('*/')
+
 
 class WhitespaceToken(Node):
-    """A <whitespace> token."""
+    """A :diagram:`whitespace-token`.
+
+    .. autoattribute:: type
+
+    """
     __slots__ = []
     type = 'whitespace'
+    repr_format = '<{self.__class__.__name__}>'
+
+    def _serialize_to(self, write):
+        write(' ')
 
 
 class LiteralToken(Node):
     r"""Token that represents one or more characters as in the CSS source.
 
+    .. autoattribute:: type
+
     .. attribute:: value
 
         A string of one to four characters.
 
-    Instances compare equal to their value.
+    Instances compare equal to their :attr:`value`,
+    so that these are equivalent:
 
-    This regroups what the spec defines as
-    <delim>, <colon>, <semicolon>, <comma>, <cdc>, <cdo>,
-    <include-match>, <dash-match>, <prefix-match>, <suffix-match>,
-    and <substring-match> tokens.
+    .. code-block:: python
+
+        if node == ';':
+        if node.type == 'literal' and node.value == ';':
+
+    This regroups what `the specification`_ defines as separate token types:
+
+    .. _the specification: http://dev.w3.org/csswg/css-syntax-3/
+
+    * *<colon-token>* ``:``
+    * *<semicolon-token>* ``;``
+    * *<comma-token>* ``,``
+    * *<cdc-token>* ``-->``
+    * *<cdo-token>* ``<!--``
+    * *<include-match-token>* ``~=``
+    * *<dash-match-token>* ``|=``
+    * *<prefix-match-token>* ``^=``
+    * *<suffix-match-token>* ``$=``
+    * *<substring-match-token>* ``*=``
+    * *<column-token>* ``||``
+    * *<delim-token>* (a single ASCII character not part of any another token)
 
     """
     __slots__ = ['value']
     type = 'literal'
+    repr_format = '<{self.__class__.__name__} {self.value}>'
 
     def __init__(self, line, column, value):
         Node.__init__(self, line, column)
@@ -106,9 +207,14 @@ class LiteralToken(Node):
     def __ne__(self, other):
         return not self == other
 
+    def _serialize_to(self, write):
+        write(self.value)
+
 
 class IdentToken(Node):
-    """A <ident> token.
+    """An :diagram:`ident-token`.
+
+    .. autoattribute:: type
 
     .. attribute:: value
 
@@ -117,67 +223,94 @@ class IdentToken(Node):
     .. attribute:: lower_value
 
         Same as :attr:`value` but normalized to *ASCII lower case*,
-        see :func:`~tinycss2.ascii_lower`.
+        see :func:`~webencodings.ascii_lower`.
         This is the value to use when comparing to a CSS keyword.
 
     """
     __slots__ = ['value', 'lower_value']
     type = 'ident'
+    repr_format = '<{self.__class__.__name__} {self.value}>'
 
     def __init__(self, line, column, value):
         Node.__init__(self, line, column)
         self.value = value
         self.lower_value = ascii_lower(value)
 
+    def _serialize_to(self, write):
+        write(serialize_identifier(self.value))
+
 
 class AtKeywordToken(Node):
-    """A <at-keyword> token.
+    """An :diagram:`at-keyword-token`.
+
+    .. autoattribute:: type
 
     .. attribute:: value
 
-        The unescaped value, as an Unicode string.
+        The unescaped value, as an Unicode string, without the preceding ``@``.
 
     .. attribute:: lower_value
 
         Same as :attr:`value` but normalized to *ASCII lower case*,
-        see :func:`~tinycss2.ascii_lower`.
+        see :func:`~webencodings.ascii_lower`.
         This is the value to use when comparing to a CSS at-keyword.
 
+        .. code-block:: python
+
+            if node.type == 'at-keyword' and node.lower_value == 'import': ...
+
     """
     __slots__ = ['value', 'lower_value']
     type = 'at-keyword'
+    repr_format = '<{self.__class__.__name__} @{self.value}>'
 
     def __init__(self, line, column, value):
         Node.__init__(self, line, column)
         self.value = value
         self.lower_value = ascii_lower(value)
 
+    def _serialize_to(self, write):
+        write('@')
+        write(serialize_identifier(self.value))
+
 
 class HashToken(Node):
-    r"""A <hash> token.
+    r"""A :diagram:`hash-token`.
+
+    .. autoattribute:: type
 
     .. attribute:: value
 
-        The unescaped value, as an Unicode string.
+        The unescaped value, as an Unicode string, without the preceding ``#``.
 
     .. attribute:: is_identifier
 
-        A :class:`bool`, true if the CSS source for this token
+        A boolean, true if the CSS source for this token
         was ``#`` followed by a valid identifier.
         (Only such hash tokens are valid ID selectors.)
 
     """
     __slots__ = ['value', 'is_identifier']
     type = 'hash'
+    repr_format = '<{self.__class__.__name__} #{self.value}>'
 
     def __init__(self, line, column, value, is_identifier):
         Node.__init__(self, line, column)
         self.value = value
         self.is_identifier = is_identifier
 
+    def _serialize_to(self, write):
+        write('#')
+        if self.is_identifier:
+            write(serialize_identifier(self.value))
+        else:
+            write(serialize_name(self.value))
+
 
 class StringToken(Node):
-    """A <string> token.
+    """A :diagram:`string-token`.
+
+    .. autoattribute:: type
 
     .. attribute:: value
 
@@ -186,14 +319,22 @@ class StringToken(Node):
     """
     __slots__ = ['value']
     type = 'string'
+    repr_format = '<{self.__class__.__name__} {self.value:r}>'
 
     def __init__(self, line, column, value):
         Node.__init__(self, line, column)
         self.value = value
 
+    def _serialize_to(self, write):
+        write('"')
+        write(serialize_string_value(self.value))
+        write('"')
+
 
 class URLToken(Node):
-    """A <url> token.
+    """An :diagram:`url-token`.
+
+    .. autoattribute:: type
 
     .. attribute:: value
 
@@ -203,31 +344,53 @@ class URLToken(Node):
     """
     __slots__ = ['value']
     type = 'url'
+    repr_format = '<{self.__class__.__name__} url({self.value})>'
 
     def __init__(self, line, column, value):
         Node.__init__(self, line, column)
         self.value = value
 
+    def _serialize_to(self, write):
+        write('url("')
+        write(serialize_string_value(self.value))
+        write('")')
+
 
 class UnicodeRangeToken(Node):
-    """An <unicode-range> token.
+    """An :diagram:`unicode-range-token`.
 
-    .. attribute:: range
+    .. autoattribute:: type
 
-        A ``(start, end)`` tuple of Unicode characters for an inclusive range,
-        or ``None`` for empty ranges.
+    .. attribute:: start
+
+        The start of the range, as an integer between 0 and 1114111.
+
+    .. attribute:: end
+
+        The end of the range, as an integer between 0 and 1114111.
+        Same as :attr:`start` if the source only specified one value.
 
     """
-    __slots__ = ['range']
+    __slots__ = ['start', 'end']
     type = 'unicode-range'
+    repr_format = '<{self.__class__.__name__} {self.start} {self.end}>'
 
-    def __init__(self, line, column, range_):
+    def __init__(self, line, column, start, end):
         Node.__init__(self, line, column)
-        self.range = range_
+        self.start = start
+        self.end = end
+
+    def _serialize_to(self, write):
+        if self.end == self.start:
+            write('U+%X' % self.start)
+        else:
+            write('U+%X-%X' % (self.start, self.end))
 
 
 class NumberToken(Node):
-    """A <number> token.
+    """A :diagram:`numer-token`.
+
+    .. autoattribute:: type
 
     .. attribute:: value
 
@@ -249,6 +412,7 @@ class NumberToken(Node):
     """
     __slots__ = ['value', 'int_value', 'is_integer', 'representation']
     type = 'number'
+    repr_format = '<{self.__class__.__name__} {self.representation}>'
 
     def __init__(self, line, column, value, int_value, representation):
         Node.__init__(self, line, column)
@@ -257,9 +421,14 @@ class NumberToken(Node):
         self.is_integer = int_value is not None
         self.representation = representation
 
+    def _serialize_to(self, write):
+        write(self.representation)
+
 
 class PercentageToken(Node):
-    """A <percentage> token.
+    """A :diagram:`percentage-token`.
+
+    .. autoattribute:: type
 
     .. attribute:: value
 
@@ -283,6 +452,7 @@ class PercentageToken(Node):
     """
     __slots__ = ['value', 'int_value', 'is_integer', 'representation']
     type = 'percentage'
+    repr_format = '<{self.__class__.__name__} {self.representation}%>'
 
     def __init__(self, line, column, value, int_value, representation):
         Node.__init__(self, line, column)
@@ -291,9 +461,15 @@ class PercentageToken(Node):
         self.is_integer = int_value is not None
         self.representation = representation
 
+    def _serialize_to(self, write):
+        write(self.representation)
+        write('%')
+
 
 class DimensionToken(Node):
-    """A <dimension> token.
+    """A :diagram:`dimension-token`.
+
+    .. autoattribute:: type
 
     .. attribute:: value
 
@@ -307,7 +483,7 @@ class DimensionToken(Node):
 
     .. attribute:: is_integer
 
-        Whether the token’value was syntactically an integer, as a boolean.
+        Whether the token’s value was syntactically an integer, as a boolean.
 
     .. attribute:: representation
 
@@ -320,14 +496,20 @@ class DimensionToken(Node):
 
     .. attribute:: lower_unit
 
-        Same as :attr:`name` but normalized to *ASCII lower case*,
-        see :func:`~tinycss2.ascii_lower`.
+        Same as :attr:`unit` but normalized to *ASCII lower case*,
+        see :func:`~webencodings.ascii_lower`.
         This is the value to use when comparing to a CSS unit.
 
+        .. code-block:: python
+
+            if node.type == 'dimension' and node.lower_unit == 'px': ...
+
     """
     __slots__ = ['value', 'int_value', 'is_integer', 'representation',
                  'unit', 'lower_unit']
     type = 'dimension'
+    repr_format = ('<{self.__class__.__name__} '
+                   '{self.representation}{self.unit}>')
 
     def __init__(self, line, column, value, int_value, representation, unit):
         Node.__init__(self, line, column)
@@ -338,60 +520,96 @@ class DimensionToken(Node):
         self.unit = unit
         self.lower_unit = ascii_lower(unit)
 
+    def _serialize_to(self, write):
+        write(self.representation)
+        # Disambiguate with scientific notation
+        unit = self.unit
+        if unit in ('e', 'E') or unit.startswith(('e-', 'E-')):
+            write('\\65 ')
+            write(serialize_name(unit[1:]))
+        else:
+            write(serialize_identifier(unit))
+
 
 class ParenthesesBlock(Node):
-    """A () block.
+    """A :diagram:`()-block`.
+
+    .. autoattribute:: type
 
     .. attribute:: content
 
-        The content of the block, as list of :ref:`component values`.
+        The content of the block, as list of :term:`component values`.
         The ``(`` and ``)`` markers themselves are not represented in the list.
 
     """
     __slots__ = ['content']
     type = '() block'
+    repr_format = '<{self.__class__.__name__} ( {self.content} )>'
 
     def __init__(self, line, column, content):
         Node.__init__(self, line, column)
         self.content = content
 
+    def _serialize_to(self, write):
+        write('(')
+        _serialize_to(self.content, write)
+        write(')')
+
 
 class SquareBracketsBlock(Node):
-    """A [] block.
+    """A :diagram:`[]-block`.
+
+    .. autoattribute:: type
 
     .. attribute:: content
 
-        The content of the block, as list of :ref:`component values`.
+        The content of the block, as list of :term:`component values`.
         The ``[`` and ``]`` markers themselves are not represented in the list.
 
     """
     __slots__ = ['content']
     type = '[] block'
+    repr_format = '<{self.__class__.__name__} [ {self.content} ]>'
 
     def __init__(self, line, column, content):
         Node.__init__(self, line, column)
         self.content = content
 
+    def _serialize_to(self, write):
+        write('[')
+        _serialize_to(self.content, write)
+        write(']')
+
 
 class CurlyBracketsBlock(Node):
-    """A {} block.
+    """A :diagram:`{}-block`.
+
+    .. autoattribute:: type
 
     .. attribute:: content
 
-        The content of the block, as list of :ref:`component values`.
+        The content of the block, as list of :term:`component values`.
         The ``[`` and ``]`` markers themselves are not represented in the list.
 
     """
     __slots__ = ['content']
     type = '{} block'
+    repr_format = '<{self.__class__.__name__} {{ {self.content} }}>'
 
     def __init__(self, line, column, content):
         Node.__init__(self, line, column)
         self.content = content
 
+    def _serialize_to(self, write):
+        write('{')
+        _serialize_to(self.content, write)
+        write('}')
 
-class Function(Node):
-    """A CSS function.
+
+class FunctionBlock(Node):
+    """A :diagram:`function-block`.
+
+    .. autoattribute:: type
 
     .. attribute:: name
 
@@ -400,12 +618,12 @@ class Function(Node):
     .. attribute:: lower_name
 
         Same as :attr:`name` but normalized to *ASCII lower case*,
-        see :func:`~tinycss2.ascii_lower`.
+        see :func:`~webencodings.ascii_lower`.
         This is the value to use when comparing to a CSS function name.
 
     .. attribute:: arguments
 
-        The arguments of the function, as list of :ref:`component values`.
+        The arguments of the function, as list of :term:`component values`.
         The ``(`` and ``)`` markers themselves are not represented in the list.
         Commas are not special, but represented as :obj:`LiteralToken` objects
         in the list.
@@ -413,6 +631,7 @@ class Function(Node):
     """
     __slots__ = ['name', 'lower_name', 'arguments']
     type = 'function'
+    repr_format = '<{self.__class__.__name__} {self.name}( {self.arguments} )>'
 
     def __init__(self, line, column, name, arguments):
         Node.__init__(self, line, column)
@@ -420,39 +639,51 @@ class Function(Node):
         self.lower_name = ascii_lower(name)
         self.arguments = arguments
 
+    def _serialize_to(self, write):
+        write(serialize_identifier(self.name))
+        write('(')
+        _serialize_to(self.arguments, write)
+        write(')')
+
 
 class Declaration(Node):
-    """A (property or descriptor) declaration.
+    """A (property or descriptor) :diagram:`declaration`.
 
-    Syntax:
-    ``<ident> <whitespace>* ':' <token>* ( '!' <ident("important")> )?``
+    .. autoattribute:: type
 
     .. attribute:: name
 
-        The unescaped value, as an Unicode string.
+        The unescaped name, as an Unicode string.
+
+    .. autoattribute:: type
 
     .. attribute:: lower_name
 
         Same as :attr:`name` but normalized to *ASCII lower case*,
-        see :func:`~tinycss2.ascii_lower`.
+        see :func:`~webencodings.ascii_lower`.
         This is the value to use when comparing to
         a CSS property or descriptor name.
 
+        .. code-block:: python
+
+            if node.type == 'declaration' and node.lower_name == 'color': ...
+
     .. attribute:: value
 
-        The declaration value as a list of :ref:`component values`:
+        The declaration value as a list of :term:`component values`:
         anything between ``:`` and
         the end of the declaration, or ``!important``.
 
     .. attribute:: important
 
-        A boolean, true if the declaration had an ``!important`` markers.
+        A boolean, true if the declaration had an ``!important`` marker.
         It is up to the consumer to reject declarations that do not accept
         this flag, such as non-property descriptor declarations.
 
     """
     __slots__ = ['name', 'lower_name', 'value', 'important']
     type = 'declaration'
+    repr_format = '<{self.__class__.__name__} {self.name}: {self.value}>'
 
     def __init__(self, line, column, name, lower_name, value, important):
         Node.__init__(self, line, column)
@@ -461,46 +692,62 @@ class Declaration(Node):
         self.value = value
         self.important = important
 
+    def _serialize_to(self, write):
+        write(serialize_identifier(self.name))
+        write(':')
+        _serialize_to(self.value, write)
+        write(';')
+
 
 class QualifiedRule(Node):
-    """A qualified rule, ie. a rule that is not an at-rule.
+    """A :diagram:`qualified rule`.
 
-    Qulified rules are often style rules,
-    where the prelude is parsed as a selector list
-    and the content as a declaration list.
+    The interpretation of style rules depend on their context.
+    At the top-level of a stylesheet
+    or in a conditional rule such as ``@media``,
+    they are style rules where the :attr:`prelude` is a list of Selectors
+    and the :attr:`body` is a list of property declarations.
 
-    Syntax:
-    ``<token except {} block>* <{} block>``
+    .. autoattribute:: type
 
     .. attribute:: prelude
 
         The rule’s prelude, the part before the {} block,
-        as a list of :ref:`component values`.
+        as a list of :term:`component values`.
 
     .. attribute:: content
 
         The rule’s content, the part inside the {} block,
-        as a list of :ref:`component values`.
+        as a list of :term:`component values`.
 
     """
     __slots__ = ['prelude', 'content']
     type = 'qualified-rule'
+    repr_format = ('<{self.__class__.__name__} '
+                   '{self.prelude} {{ {self.content} }}>')
 
     def __init__(self, line, column, prelude, content):
         Node.__init__(self, line, column)
         self.prelude = prelude
         self.content = content
 
+    def _serialize_to(self, write):
+        _serialize_to(self.prelude, write)
+        write('{')
+        _serialize_to(self.content, write)
+        write('}')
+
 
 class AtRule(Node):
-    """An at-rule.
+    """An :diagram:`at-rule`.
 
-    The interpretation of at-rules depend on their at-keywords.
-    Most at-rules (ie. at-keyword values) are only allowed in some context,
+    The interpretation of at-rules depend on their at-keyword
+    as well as their context.
+    Most types of at-rules (ie. at-keyword values)
+    are only allowed in some context,
     and must either end with a {} block or a semicolon.
 
-    Syntax:
-    ``<at-keyword> <token except {} block>* ( <{} block> | ';' )``
+    .. autoattribute:: type
 
     .. attribute:: at_keyword
 
@@ -510,24 +757,30 @@ class AtRule(Node):
     .. attribute:: lower_at_keyword
 
         Same as :attr:`at_keyword` but normalized to *ASCII lower case*,
-        see :func:`~tinycss2.ascii_lower`.
+        see :func:`~webencodings.ascii_lower`.
         This is the value to use when comparing to a CSS at-keyword.
 
+        .. code-block:: python
+
+            if node.type == 'at-rule' and node.lower_at_keyword == 'import': ...
+
     .. attribute:: prelude
 
         The rule’s prelude, the part before the {} block or semicolon,
-        as a list of :ref:`component values`.
+        as a list of :term:`component values`.
 
     .. attribute:: content
 
         The rule’s content, if any.
-        The block’s content as a list of :ref:`component values`
+        The block’s content as a list of :term:`component values`
         for at-rules with a {} block,
         or :obj:`None` for at-rules ending with a semicolon.
 
     """
     __slots__ = ['at_keyword', 'lower_at_keyword', 'prelude', 'content']
     type = 'at-rule'
+    repr_format = ('<{self.__class__.__name__} '
+                   '@{self.at_keyword} {self.prelude} {{ {self.content} }}>')
 
     def __init__(self, line, column,
                  at_keyword, lower_at_keyword, prelude, content):
@@ -536,3 +789,14 @@ class AtRule(Node):
         self.lower_at_keyword = lower_at_keyword
         self.prelude = prelude
         self.content = content
+
+    def _serialize_to(self, write):
+        write('@')
+        write(serialize_identifier(self.at_keyword))
+        _serialize_to(self.prelude, write)
+        if self.content is None:
+            write(';')
+        else:
+            write('{')
+            _serialize_to(self.content, write)
+            write('}')
